/*
 * Copyright (c) 2001-2002 by David Brownell
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifdef __KERNEL__

/* This file contains declarations of usbcore internals that are mostly
 * used or exposed by Host Controller Drivers.
 */

/*
 * USB Packet IDs (PIDs)
 */
#define USB_PID_UNDEF_0			0xf0
#define USB_PID_OUT			0xe1
#define USB_PID_ACK			0xd2
#define USB_PID_DATA0			0xc3
#define USB_PID_PING			0xb4	/* USB 2.0 */
#define USB_PID_SOF			0xa5
#define USB_PID_NYET			0x96	/* USB 2.0 */
#define USB_PID_DATA2			0x87	/* USB 2.0 */
#define USB_PID_SPLIT			0x78	/* USB 2.0 */
#define USB_PID_IN			0x69
#define USB_PID_NAK			0x5a
#define USB_PID_DATA1			0x4b
#define USB_PID_PREAMBLE		0x3c	/* Token mode */
#define USB_PID_ERR			0x3c	/* USB 2.0: handshake mode */
#define USB_PID_SETUP			0x2d
#define USB_PID_STALL			0x1e
#define USB_PID_MDATA			0x0f	/* USB 2.0 */

/*-------------------------------------------------------------------------*/

/*
 * USB Host Controller Driver (usb_hcd) framework
 *
 * Since "struct usb_bus" is so thin, you can't share much code in it.
 * This framework is a layer over that, and should be more sharable.
 */

/*-------------------------------------------------------------------------*/

struct usb_hcd {	/* usb_bus.hcpriv points to this */

 /*
  * housekeeping
  */
 struct usb_bus		self;		/* hcd is-a bus */

 #ifdef USB_DBGSTRING_ON
	const char		*product_desc;	/* product/vendor string */
 #endif // USB_DBGSTRING_ON
 #ifdef USB_DBGSTRING_ON
	const char		*description;	/* "ehci-hcd" etc */
 #endif // USB_DBGSTRING_ON

 #ifndef NO_TIMER
	struct timer_list	rh_timer;	/* drives root hub */
  #else
	struct urb * rh_urb ;
 #endif // ndef NO_TIMER
 struct list_head	dev_list;	/* devices on this bus */
 #ifndef NO_WORKQUEUE
	struct work_struct	work;
 #endif // NO_WORKQUEUE

 /*
  * hardware info/state
  */
 struct hc_driver	*driver;	/* hw-specific hooks */
#ifdef _ETCPIP
#if _USB2 == 1	// VIA vt6212
 int			dev_idx_for_polling;	/* device index for polling */
#endif
#endif	// defined(_ETCPIP)
 int			irq;		/* irq allocated */
 void			*regs;		/* device memory/io */
 struct device		*controller;	/* handle to hardware */

 /* a few non-PCI controllers exist, mostly for OHCI */
 
 #ifdef CONFIG_PCI
 	struct pci_dev		*pdev;		/* pci is typical */
	int			region;		/* pci region for regs */
	u32			pci_state [16];	/* for PM state save */
	/*
	#ifndef _ETCPIP
	*/
		atomic_t		resume_count;	/* multiple resumes issue */
	/*
	 #else
		int resume_count ;
	#endif
	*/
 #endif

 #ifndef USB_NO_DMA_POOL // 2003/7/2 B. S. Liaw
	#define HCD_BUFFER_POOLS	4
 
	 #ifdef CONFIG_PCI
		struct pci_pool		*pool [HCD_BUFFER_POOLS];
	  #else
		struct dma_pool		*pool [HCD_BUFFER_POOLS];
	 #endif
 #endif // ndef USB_NO_DMA_POOL
 
	int			state;
#	define	__ACTIVE		0x01
#	define	__SLEEPY		0x02
#	define	__SUSPEND		0x04
#	define	__TRANSIENT		0x80

#	define	USB_STATE_HALT		0
#	define	USB_STATE_RUNNING	(__ACTIVE)
#	define	USB_STATE_READY		(__ACTIVE|__SLEEPY)
#	define	USB_STATE_QUIESCING	(__SUSPEND|__TRANSIENT|__ACTIVE)
#	define	USB_STATE_RESUMING	(__SUSPEND|__TRANSIENT)
#	define	USB_STATE_SUSPENDED	(__SUSPEND)

#define	HCD_IS_RUNNING(state) ((state) & __ACTIVE)
#define	HCD_IS_SUSPENDED(state) ((state) & __SUSPEND)

	/* more shared queuing code would be good; it should support
	 * smarter scheduling, handle transaction translators, etc;
	 * input size of periodic table to an interrupt scheduler. 
	 * (ohci 32, uhci 1024, ehci 256/512/1024).
	 */
 /*
 #ifdef NO_SYS_FS
	struct usb_bus * bus;    
 #endif // NO_SYS_FS
 */
 
};

/* 2.4 does this a bit differently ... */
static inline struct usb_bus *hcd_to_bus (struct usb_hcd *hcd)
{
	return &hcd->self;
}
/*
#define USB_HCD_EP_MAX 32 // 2003/7/4 B. S. Liaw 
							and move to UsbCfg.h
*/
struct hcd_dev {	/* usb_device.hcpriv points to this */
 struct list_head	dev_list;	/* on this hcd */
 struct list_head	urb_list;	/* pending on this dev */

 /* per-configuration HC/HCD state, such as QH or ED */
 // #ifndef USB_GLOBAL_EP // 2003/7/15 B. S. Liaw
	int ep_inited ; // 2003/7/3 B. S. Liaw
	/*
	void	*ep[32];
	*/
	void	*ep[USB_HCD_EP_MAX] ; // 2003/7/3 B. S. Liaw
 // #endif // ndef USB_GLOBAL_EP
};

// urb.hcpriv is really hardware-specific
/*
#ifndef _ETCPIP
	struct hcd_timeout {	// timeouts we allocate 
		struct list_head	timeout_list;
		struct timer_list	timer;
	};
#endif
*/

/*-------------------------------------------------------------------------*/

/*
 * FIXME usb_operations should vanish or become hc_driver,
 * when usb_bus and usb_hcd become the same thing.
 */

struct usb_operations {
 int (*allocate)(struct usb_device *);
 int (*deallocate)(struct usb_device *);
 int (*get_frame_number) (struct usb_device *usb_dev);
 #ifndef _ETCPIP
	int (*submit_urb) (struct urb *urb, int mem_flags);
  #else
	int (*submit_urb) (struct urb *urb);
 #endif
 int (*unlink_urb) (struct urb *urb);

 /* allocate dma-consistent buffer for URB_DMA_NOMAPPING */
 void *(*buffer_alloc)(struct usb_bus *bus, size_t size,
			int mem_flags,
			dma_addr_t *dma);
 void (*buffer_free)(struct usb_bus *bus, size_t size,
			void *addr, dma_addr_t dma);
};

/* each driver provides one of these, and hardware init support */

struct pt_regs;

struct hc_driver {
 #ifdef USB_DBGSTRING_ON
	const char	*description;	/* "ehci-hcd" etc */
 #endif // USB_DBGSTRING_ON

 /* irq handler */
 #ifndef _ETCPIP
	void	(*irq) (struct usb_hcd *hcd, struct pt_regs *regs);
  #else
	// void	(*irq) (void);
 #endif

 int	flags;
 #define	HCD_MEMORY	0x0001		/* HC regs use memory (else I/O) */
 #define	HCD_USB11	0x0010		/* USB 1.1 */
 #define	HCD_USB2	0x0020		/* USB 2.0 */

	/* called to init HCD and root hub */
	int	(*start) (struct usb_hcd *hcd);

	/* called after all devices were suspended */
	int	(*suspend) (struct usb_hcd *hcd, u32 state);

	/* called before any devices get resumed */
	int	(*resume) (struct usb_hcd *hcd);

	/* cleanly make HCD stop writing memory and doing I/O */
	void	(*stop) (struct usb_hcd *hcd);

	/* return current frame number */
	int	(*get_frame_number) (struct usb_hcd *hcd);

	/* memory lifecycle */
	struct usb_hcd	*(*hcd_alloc) (void);
	void		(*hcd_free) (struct usb_hcd *hcd);

	/* manage i/o requests, device state */
 #ifndef _ETCPIP
	int	(*urb_enqueue) (struct usb_hcd *hcd, struct urb *urb,
					int mem_flags);
  #else
	int	(*urb_enqueue) (struct usb_hcd *hcd, struct urb *urb) ;
 #endif
 
 #ifdef USB_NO_ISR_FREE
	void (*urb_free_priv) (struct usb_hcd *hcd,
    						void *hcpriv) ;
    /*
	void (*urb_free_priv) (struct usb_hcd *hcd,
                           struct urb *the_urb
                          ) ;
    */
 #endif // USB_NO_ISR_FREE

 int	(*urb_dequeue) (struct usb_hcd *hcd, struct urb *urb);

 // frees configuration resources -- allocated as needed during
 // urb_enqueue, and not freed by urb_dequeue
 /*
 void	(*free_config) (struct usb_hcd *hcd,
				struct usb_device *dev);
 */
 /* root hub support */
 int		(*hub_status_data) (struct usb_hcd *hcd, char *buf);
 int		(*hub_control) (struct usb_hcd *hcd,
				u16 typeReq, u16 wValue, u16 wIndex,
				char *buf, u16 wLength);
};
/*
#ifndef _ETCPIP
	extern void usb_hcd_giveback_urb (struct usb_hcd *hcd, struct urb *urb, struct pt_regs *regs);
 #else
*/
	extern void usb_hcd_giveback_urb (struct usb_hcd *hcd, struct urb *urb);
/*
#endif
*/

extern void usb_bus_init (struct usb_bus *bus);

extern void usb_rh_status_dequeue (struct usb_hcd *hcd, struct urb *urb);

#ifdef CONFIG_PCI
	struct pci_dev;
	struct pci_device_id;
#endif // CONFIG_PCI

#ifndef _ETCPIP
	extern int usb_hcd_pci_probe (struct pci_dev *dev,
    		const struct pci_device_id *id ) ;
#elif defined(CONFIG_PCI)
	extern int usb_hcd_pci_probe (struct pci_dev *dev,
			const struct pci_device_id *id, int irq, void *mem_base) ;
#else
	extern int usb_hcd_probe(struct device *dev_dev , int irq,
    						void *mem_base,
                            const struct hc_driver *the_hc_driver ) ;
#endif

#ifdef CONFIG_PCI
	extern void usb_hcd_pci_remove (struct pci_dev *dev);
#endif // CONFIG_PCI

#ifdef CONFIG_PCI
	#ifdef CONFIG_PM
	// FIXME:  see Documentation/power/pci.txt (2.4.6 and later?)
	// extern int usb_hcd_pci_save_state (struct pci_dev *dev, u32 state);
	extern int usb_hcd_pci_suspend (struct pci_dev *dev, u32 state);
	extern int usb_hcd_pci_resume (struct pci_dev *dev);
	// extern int usb_hcd_pci_enable_wake (struct pci_dev *dev, u32 state, int flg);
	#endif /* CONFIG_PM */

#endif // CONFIG_PCI

/* pci-ish (pdev null is ok) buffer alloc/mapping support */
int hcd_buffer_create (struct usb_hcd *hcd);
void hcd_buffer_destroy (struct usb_hcd *hcd);

void *hcd_buffer_alloc (struct usb_bus *bus, size_t size,
	int mem_flags, dma_addr_t *dma);
// #ifndef _ETCPIP
	void hcd_buffer_free (struct usb_bus *bus, size_t size,
		void *addr, dma_addr_t dma);
// #endif // ndef _ETCPIP

/* generic bus glue, needed for host controllers that don't use PCI */

extern struct usb_operations usb_hcd_operations;

#ifndef _ETCPIP
	extern void usb_hcd_irq (int irq, void *__hcd, struct pt_regs *r);
 #else
	extern void usb_hcd_irq (int irq, void *__hcd);
#endif

extern void usb_hc_died (struct usb_hcd *hcd);

/* -------------------------------------------------------------------------- */

/* Enumeration is only for the hub driver, or HCD virtual root hubs */
extern int usb_new_device(struct usb_device *dev, struct device *parent);
extern void usb_connect(struct usb_device *dev);
extern void usb_disconnect(struct usb_device **);

/* exported to hub driver ONLY to support usb_reset_device () */
extern int usb_get_configuration(struct usb_device *dev);
extern void usb_set_maxpacket(struct usb_device *dev);
extern void usb_destroy_configuration(struct usb_device *dev);
extern int usb_set_address(struct usb_device *dev);

/* use these only before the device's address has been set */
#define usb_snddefctrl(dev)		((PIPE_CONTROL << 30))
#define usb_rcvdefctrl(dev)		((PIPE_CONTROL << 30) | USB_DIR_IN)

/*-------------------------------------------------------------------------*/

/*
 * HCD Root Hub support
 */

#include "hub.h"

/* (shifted) direction/type/recipient from the USB 2.0 spec, table 9.2 */
#define DeviceRequest \
	((USB_DIR_IN|USB_TYPE_STANDARD|USB_RECIP_DEVICE)<<8)
#define DeviceOutRequest \
	((USB_DIR_OUT|USB_TYPE_STANDARD|USB_RECIP_DEVICE)<<8)

#define InterfaceRequest \
	((USB_DIR_IN|USB_TYPE_STANDARD|USB_RECIP_INTERFACE)<<8)

#define EndpointRequest \
	((USB_DIR_IN|USB_TYPE_STANDARD|USB_RECIP_INTERFACE)<<8)
#define EndpointOutRequest \
	((USB_DIR_OUT|USB_TYPE_STANDARD|USB_RECIP_INTERFACE)<<8)

/* table 9.6 standard features */
#define DEVICE_REMOTE_WAKEUP	1
#define ENDPOINT_HALT		0

/* class requests from the USB 2.0 hub spec, table 11-15 */
/* GetBusState and SetHubDescriptor are optional, omitted */
#define ClearHubFeature		(0x2000 | USB_REQ_CLEAR_FEATURE)
#define ClearPortFeature	(0x2300 | USB_REQ_CLEAR_FEATURE)
#define GetHubDescriptor	(0xa000 | USB_REQ_GET_DESCRIPTOR)
#define GetHubStatus		(0xa000 | USB_REQ_GET_STATUS)
#define GetPortStatus		(0xa300 | USB_REQ_GET_STATUS)
#define SetHubFeature		(0x2000 | USB_REQ_SET_FEATURE)
#define SetPortFeature		(0x2300 | USB_REQ_SET_FEATURE)


/*-------------------------------------------------------------------------*/

/*
 * Generic bandwidth allocation constants/support
 */
#define FRAME_TIME_USECS	1000L
#define BitTime(bytecount)  (7 * 8 * bytecount / 6)  /* with integer truncation */
		/* Trying not to use worst-case bit-stuffing
                   of (7/6 * 8 * bytecount) = 9.33 * bytecount */
		/* bytecount = data payload byte count */

#define NS_TO_US(ns)	((ns + 500L) / 1000L)
			/* convert & round nanoseconds to microseconds */

extern void usb_claim_bandwidth (struct usb_device *dev, struct urb *urb,
		int bustime, int isoc);
extern void usb_release_bandwidth (struct usb_device *dev, struct urb *urb,
		int isoc);

/*
 * Full/low speed bandwidth allocation constants/support.
 */
#define BW_HOST_DELAY	1000L		/* nanoseconds */
#define BW_HUB_LS_SETUP	333L		/* nanoseconds */
                        /* 4 full-speed bit times (est.) */

#define FRAME_TIME_BITS         12000L		/* frame = 1 millisecond */
#define FRAME_TIME_MAX_BITS_ALLOC	(90L * FRAME_TIME_BITS / 100L)
#define FRAME_TIME_MAX_USECS_ALLOC	(90L * FRAME_TIME_USECS / 100L)

extern int usb_check_bandwidth (struct usb_device *dev, struct urb *urb);

/*
 * Ceiling microseconds (typical) for that many bytes at high speed
 * ISO is a bit less, no ACK ... from USB 2.0 spec, 5.11.3 (and needed
 * to preallocate bandwidth)
 */
#define USB2_HOST_DELAY	5	/* nsec, guess */
#define HS_USECS(bytes) NS_TO_US ( ((55 * 8 * 2083)/1000) \
	+ ((2083UL * (3167 + BitTime (bytes)))/1000) \
	+ USB2_HOST_DELAY)
#define HS_USECS_ISO(bytes) NS_TO_US ( ((long)(38 * 8 * 2.083)) \
	+ ((2083UL * (3167 + BitTime (bytes)))/1000) \
	+ USB2_HOST_DELAY)

extern long usb_calc_bus_time (int speed, int is_input,
			int isoc, int bytecount);

/*-------------------------------------------------------------------------*/

extern struct usb_bus *usb_alloc_bus (struct usb_operations *);
extern void usb_free_bus (struct usb_bus *);

extern void usb_register_bus (struct usb_bus *);
extern void usb_deregister_bus (struct usb_bus *);

extern int usb_register_root_hub (struct usb_device *usb_dev,
		struct device *parent_dev);

/* for portability to 2.4, hcds should call this */
static inline int hcd_register_root (struct usb_hcd *hcd)
{
 return usb_register_root_hub (
		hcd_to_bus (hcd)->root_hub, hcd->controller);
}

/*-------------------------------------------------------------------------*/

/* exported only within usbcore */

extern struct list_head usb_bus_list;
#ifndef NO_SEMPHORE
	extern struct semaphore usb_bus_list_lock;
#endif // ndef NO_SEMPHORE

extern void usb_bus_get (struct usb_bus *bus);
extern void usb_bus_put (struct usb_bus *bus);

#ifndef NO_SYS_FS
	extern int usb_find_interface_driver (struct usb_device *dev,
		struct usb_interface *interface);
 #else
	extern int  usb_find_interface_driver(struct usb_device *,
    	unsigned int);
#endif

#define usb_endpoint_halt(dev, ep, out) ((dev)->halted[out] |= (1 << (ep)))

#define usb_endpoint_out(ep_dir)	(!((ep_dir) & USB_DIR_IN))

/*
 * USB device fs stuff
 */

#ifdef CONFIG_USB_DEVICEFS

	/*
	 * these are expected to be called from the USB core/hub thread
	 * with the kernel lock held
	 */
	extern void usbfs_add_bus(struct usb_bus *bus);
	extern void usbfs_remove_bus(struct usb_bus *bus);
	extern void usbfs_add_device(struct usb_device *dev);
	extern void usbfs_remove_device(struct usb_device *dev);
	extern void usbfs_update_special (void);

	extern int usbfs_init(void);
	extern void usbfs_cleanup(void);

 #else /* CONFIG_USB_DEVICEFS */

	static inline void usbfs_add_bus(struct usb_bus *bus) {}
	static inline void usbfs_remove_bus(struct usb_bus *bus) {}
	static inline void usbfs_add_device(struct usb_device *dev) {}
	static inline void usbfs_remove_device(struct usb_device *dev) {}
	static inline void usbfs_update_special (void) {}
	
	static inline int usbfs_init(void) { return 0; }
	static inline void usbfs_cleanup(void) { }

#endif /* CONFIG_USB_DEVICEFS */

/*-------------------------------------------------------------------------*/

/* hub.h ... DeviceRemovable in 2.4.2-ac11, gone in 2.4.10 */
// bleech -- resurfaced in 2.4.11 or 2.4.12
#define bitmap 	DeviceRemovable


/*-------------------------------------------------------------------------*/

/* random stuff */

#define	RUN_CONTEXT (in_irq () ? "in_irq" \
		: (in_interrupt () ? "in_interrupt" : "can sleep"))


#endif /* __KERNEL__ */


// #ifdef _ETCPIP
/*
#ifdef USB_OHCI
*/
extern void localhc_irq(void) ;
/*
#elif defined(USB_SOTA)   // Sota USB 1.1 Host
		extern void localhc_irq(void) ;
#endif 
*/
#ifdef SOTA_SOFT_ISR   // 2004/1/7 B. S. Liaw
	extern void sota_soft_isr(void) ;
	#ifndef USB_HC_WAIT_BY_EVENT
		extern volatile int  hc_wdh ;
	#endif // ndef USB_HC_WAIT_BY_EVENT
#endif // SOTA_SOFT_ISR

extern void localhc_free_config (struct usb_hcd *hcd,
				struct usb_device *usb_dev) ;
// #endif // _ETCPIP

extern int hcd_init_ok ;


