/* Driver for USB Mass Storage compliant devices
 *
 * $Id: transport.c,v 1.47 2002/04/22 03:39:43 mdharm Exp $
 *
 * Current development and maintenance by:
 *   (c) 1999-2002 Matthew Dharm (mdharm-usb@one-eyed-alien.net)
 *
 * Developed with the assistance of:
 *   (c) 2000 David L. Brown, Jr. (usb-storage@davidb.org)
 *   (c) 2000 Stephen J. Gowdy (SGowdy@lbl.gov)
 *   (c) 2002 Alan Stern <stern@rowland.org>
 *
 * Initial work by:
 *   (c) 1999 Michael Gee (michael@linuxspecific.com)
 *
 * This driver is based on the 'USB Mass Storage Class' document. This
 * describes in detail the protocol used to communicate with such
 * devices.  Clearly, the designers had SCSI and ATAPI commands in
 * mind when they created this document.  The commands are all very
 * similar to commands in the SCSI-II and ATAPI specifications.
 *
 * It is important to note that in a number of cases this class
 * exhibits class-specific exemptions from the USB specification.
 * Notably the usage of NAK, STALL and ACK differs from the norm, in
 * that they are used to communicate wait, failed and OK on commands.
 *
 * Also, for certain devices, the interrupt endpoint is used to convey
 * status of a command.
 *
 * Please see http://www.one-eyed-alien.net/~mdharm/linux-usb for more
 * information about this driver.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

// #ifdef _ETCPIP
#define __KERNEL__
#ifdef _NEWLIB
	#include "string.h"
#endif
#include "EtcpipDbg.h" // 2002/10/25 B. S. Liaw
          // for deciding whether to compile
          //  Debug Version or Release Version
#include "etcpip.h"
#include "UsbDbg.h"
#include "PoList.h" // osport_list.h
#include "UsbCfg.h" // Config and Setting
#include "UsbMsCfg.h" // Config and Setting for Mass Storage Class
#include "UsbMsDbg.h"
#include "UsOsport.h" // "usb_osport.h"
#ifndef NO_WAITQUEUE
	#include  "PoWait.h" // osport_wait.h
#endif // ndef NO_WAITQUEUE
// #endif // _ETCPIP

/*
#include <linux/config.h>
#include "transport.h"
*/
#include "ms_transport.h"
/*
#include "protocol.h"
*/
#include "ms_protocol.h"
/*
#include "scsiglue.h"
*/
#include "ms_scsiglue.h"
/*
#include "usb.h"
*/
#include "ms_usb.h"
/*
#include "debug.h"
*/
#include "ms_debug.h"
/*
#include <linux/sched.h>
#include <linux/errno.h>
*/
#include "PoErrno.h" // osport_errno.h
/*
#include <linux/slab.h>
*/


/***********************************************************************
 * Data transfer routines
 ***********************************************************************/

/*
 * This is subtle, so pay attention:
 * ---------------------------------
 * We're very concerned about races with a command abort.  Hanging this code
 * is a sure fire way to hang the kernel.  (Note that this discussion applies
 * only to transactions resulting from a scsi queued-command, since only
 * these transactions are subject to a scsi abort.  Other transactions, such
 * as those occurring during device-specific initialization, must be handled
 * by a separate code path.)
 *
 * The abort function (usb_storage_command_abort() in scsiglue.c) first
 * sets the machine state and the ABORTING bit in us->flags to prevent
 * new URBs from being submitted.  It then calls usb_stor_stop_transport()
 * below, which atomically tests-and-clears the URB_ACTIVE bit in us->flags
 * to see if the current_urb needs to be stopped.  Likewise, the SG_ACTIVE
 * bit is tested to see if the current_sg scatter-gather request needs to be
 * stopped.  The timeout callback routine does much the same thing.
 *
 * When a disconnect occurs, the DISCONNECTING bit in us->flags is set to
 * prevent new URBs from being submitted, and usb_stor_stop_transport() is
 * called to stop any ongoing requests.
 *
 * The submit function first verifies that the submitting is allowed
 * (neither ABORTING nor DISCONNECTING bits are set) and that the submit
 * completes without errors, and only then sets the URB_ACTIVE bit.  This
 * prevents the stop_transport() function from trying to cancel the URB
 * while the submit call is underway.  Next, the submit function must test
 * the flags to see if an abort or disconnect occurred during the submission
 * or before the URB_ACTIVE bit was set.  If so, it's essential to cancel
 * the URB if it hasn't been cancelled already (i.e., if the URB_ACTIVE bit
 * is still set).  Either way, the function must then wait for the URB to
 * finish.  Note that because the URB_ASYNC_UNLINK flag is set, the URB can
 * still be in progress even after a call to usb_unlink_urb() returns.
 *
 * The idea is that (1) once the ABORTING or DISCONNECTING bit is set,
 * either the stop_transport() function or the submitting function
 * is guaranteed to call usb_unlink_urb() for an active URB,
 * and (2) test_and_clear_bit() prevents usb_unlink_urb() from being
 * called more than once or from being called during usb_submit_urb().
 */

/* This is the completion handler which will wake us up when an URB
 * completes.
 */
/*
static void usb_stor_blocking_completion(struct urb *urb, struct pt_regs *regs)
*/
static void usb_stor_blocking_completion(struct urb *urb)
{
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
	struct completion *urb_done_ptr = (struct completion *)urb->context;

	complete(urb_done_ptr);
 #endif // ndef _ETCPIP
}
 
/* This is the timeout handler which will cancel an URB when its timeout
 * expires.
 */
static void timeout_handler(unsigned long us_)
{
	struct us_data *us = (struct us_data *) us_;

	if (__test_and_clear_bit(US_FLIDX_URB_ACTIVE, &us->flags)) {
		US_DEBUGP("Timeout -- cancelling URB\n");
		usb_unlink_urb(us->current_urb);
	}
}

/* This is the common part of the URB message submission code
 *
 * All URBs from the usb-storage driver involved in handling a queued scsi
 * command _must_ pass through this function (or something like it) for the
 * abort mechanisms to work properly.
 */
/*
static int usb_stor_msg_common(struct us_data *us, int timeout)
*/
static int usb_stor_msg_common(struct us_data *us, unsigned long timeout)
{
 /*
 #ifndef _ETCPIP
    // struct completion in <linux/completion.h>
	struct completion urb_done;
 #endif // ndef _ETCPIP
 */
 /*
 #ifndef NO_TIMER
	struct timer_list to_timer;
 #endif // ndef NO_TIMER
 */
 int status;
 // #ifdef _ETCPIP
 int length = 0 ;
 // #endif // def _ETCPIP


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_msg_common() Begin .. \n"));

 // USBMS_DBG_URB(iprintf("[USBMS URB] usb_stor_msg_common() Begin ..\n"));
    
 #ifdef USBMS_BS_ASSERT
	if ( !(USB_IS_UN_CACHED(us)) ) {
		iprintf("[ASSERT][USBMS] usb_stor_msg_common() : " ) ;
	    iprintf("Error : Not Un-Cached !!\n" ) ;
	    iprintf("[ASSERT][USBMS] us addr=%08XH \n", us ) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT

 #ifdef USBMS_BS_SYSCAL_ASSERT
	if (!(  mt_busy == 0
       ) ) {
		iprintf("[ASSERT][USBMS] usb_stor_msg_common() : " ) ;
	    iprintf("Error : mt_busy not zero !! %d \n",mt_busy) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_SYSCAL_ASSERT

 #ifdef USBMS_BS_ASSERT
	if (!( timeout > 0
	   ) ) {
		iprintf("[ASSERT][USBMS] usb_stor_msg_common() : " ) ;
	    iprintf("Error : invalid timeout=%d !! %d \n",timeout) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT

 /* don't submit URBs during abort/disconnect processing */
 if (us->flags & DONT_SUBMIT) {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_stor_msg_common(1) : DONT_SUBMIT \n" )) ;
	return -EIO;
   }

 /* set up data structures for the wakeup system */
 /*
 #ifndef _ETCPIP
	init_completion(&urb_done);
 #endif // ndef _ETCPIP
 */
 /* fill the common fields in the URB */
 /*
 #ifndef _ETCPIP
	us->current_urb->context = &urb_done;
 #endif // ndef _ETCPIP
 */
 us->current_urb->actual_length = 0;
 us->current_urb->error_count = 0;
 us->current_urb->status = 0;

 /* we assume that if transfer_buffer isn't us->iobuf then it
  * hasn't been mapped for DMA.  Yes, this is clunky, but it's
  * easier than always having the caller tell us whether the
  * transfer buffer has already been mapped. */

 us->current_urb->transfer_flags =
			/* URB_NO_SETUP_DMA_MAP only on newer 2.6.0-test7
			URB_ASYNC_UNLINK | URB_NO_SETUP_DMA_MAP;
			*/
			URB_ASYNC_UNLINK ;

 /* URB_NO_SETUP_DMA_MAP only on newer 2.6.0-test7
 if (us->current_urb->transfer_buffer == us->iobuf)
		us->current_urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
 */
 /*
 #ifndef USB_NO_VMM // vitrtul memory management
	us->current_urb->transfer_dma = us->iobuf_dma;
 #endif // ndef USB_NO_VMM
 */
 /*
 #ifndef _ETCPIP
	us->current_urb->setup_dma = us->cr_dma;
 #endif // ndef _ETCPIP
 */

 /* submit the URB */
 /*
 status = usb_submit_urb(us->current_urb, GFP_NOIO);
 status = usb_submit_urb(us->current_urb) ;
 */

 
 // #ifdef _ETCPIP
 /*
 status = usb_control_msg(usblp->usb_dev,
		dir ? usb_rcvctrlpipe(usblp->usb_dev, 0)
            : usb_sndctrlpipe(usblp->usb_dev, 0),
		// request, USB_TYPE_CLASS | dir | recip, value, usblp->ifnum,
        // buf, len, HZ * 5);
		request, type | dir | recip, value, usblp->ifnum,
        buf, len, USBLPT_CTLMSG_TIMEOUT);
        
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] usb_stor_msg_common() : rq:%02XH dir:%d recip:%d "
	"value:%d len:%d result:%d \n",
	request, !!dir, recip, value, len, retval
    ));
 */
 #ifdef USBMS_BS_ASSERT
    if  ( !( length == 0
        )  ) {
         iprintf("[ASSERT][USBMS] usb_stor_msg_common() : " ) ;
         iprintf("Error : length Not 0 !!\n" ) ;
         iprintf("[ASSERT][USBMS] Continue ....\n") ;
        }
 #endif // USBMS_BS_ASSERT
 
 USBMS_DBG_URB(iprintf(
    "[USBMS URB] usb_stor_msg_common() length value = %d\n",
    length));
    
 // This Donot free urb
 /*
 status = usb_start_wait_urb_nf(us->current_urb, timeout, 0 , &length);
 */
 status = usb_start_wait_urb_nf(us->current_urb, timeout, // 1 , 
								&length);

 #ifdef USBMS_DBG_URB_BUFFER_ON
	BLOCK_PREEMPTION;
	if (us->current_urb->transfer_buffer) {
	    unsigned char *buf_ptr ;

		buf_ptr =	(unsigned char *)(us->current_urb->transfer_buffer) ;

		USBMS_DBG_BUFFER(iprintf(
			"[USBMS BUFFER] usb_stor_msg_common() : Dump urb->transfer_buffer "
			"Leading 32 Bytes : \n")) ;
		usb_DumpMem((char *) &(buf_ptr[0]), 32 ,
					USB_DUMP_DEFAULT | USB_DUMP_ASCII);
	   }
	 else
	   {
		USBMS_DBG_BUFFER(iprintf(
			"[USBMS BUFFER] usb_stor_msg_common() : No urb->transfer_buffer to Dump !! \n")) ;
	   }
	UNBLOCK_PREEMPTION;
 #endif // USBMS_DBG_URB_BUFFER_ON

 /*
 // so, now free urb
 USBMS_DBG_URB(iprintf("[USBMS URB] usb_stor_msg_common() Call usb_free_urb() ...\n")) ;
 usb_free_urb(us->current_urb);
 */
 // #endif // _ETCPIP

 /*
 #ifdef _ETCPIP
 */
	if (status < 0) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_msg_common(2) : Error : status=%d \n" ,
			status )) ;
		dprintf("[USBMS] Error : USB I/O Falied , status=%d !! \n" , status ) ;
	    USBMS_DBG_FLOW(iprintf(
			"[USBMS FLOW] usb_stor_msg_common() End with Fail.\n"
			));
		return status;
	   }
	  else {
		// USBMS_DBG_URB(iprintf("[USBMS URB] usb_stor_msg_common() : length = %d \n" , length )) ;
		/*
	    USBMS_DBG_URB(iprintf(
	        "[USBMS URB] usb_stor_msg_common() End , length =%d \n",
	        length));
		return length;
		*/
	   }
 /*
  #else
	 if (status) {
		// something went wrong
		return status;
		}
 #endif // ndef _ETCPIP
 */

 /* since the URB has been submitted successfully, it's now okay
  * to cancel it */

 set_bit(US_FLIDX_URB_ACTIVE, &us->flags);

 /* did an abort/disconnect occur during the submission? */
 if (us->flags & DONT_SUBMIT) {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_stor_msg_common(3) : DONT_SUBMIT \n" )) ;
	/* cancel the URB, if it hasn't been cancelled already */
	USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_stor_msg_common(4) : Not Call usb_unlink_urb() !! \n" )) ;
	return -ENODEV ;
	/*
	if (__test_and_clear_bit(US_FLIDX_URB_ACTIVE, &us->flags)) {
		US_DEBUGP("-- cancelling URB\n");
		USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_stor_msg_common(5) : usb_unlink_urb() .. \n" )) ;
		usb_unlink_urb(us->current_urb);
	   }
	*/
   }

 /* submit the timeout timer, if a timeout was requested */
 /*
 #ifndef NO_TIMER
	if (timeout > 0) {
		init_timer(&to_timer);
		to_timer.expires = jiffies + timeout;
		to_timer.function = timeout_handler;
		to_timer.data = (unsigned long) us;
		add_timer(&to_timer);
	}
 #endif // ndef NO_TIMER
 */
 /* wait for the completion of the URB */
 /*
 #ifndef _ETCPIP
    // wait_for_completion() is in <linux/completion.h>
	wait_for_completion(&urb_done);
 #endif // ndef _ETCPIP
 */
 clear_bit(US_FLIDX_URB_ACTIVE, &us->flags);
 
 /* clean up the timeout timer */
 /*
 #ifndef NO_TIMER
	if (timeout > 0)
		del_timer_sync(&to_timer);
 #endif // ndef NO_TIMER
 */
 /* return the URB status */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_msg_common() End.\n"));
 return us->current_urb->status;
}

/*
 * Transfer one control message, with timeouts, and allowing early
 * termination.  Return codes are usual -Exxx, *not* USB_STOR_XFER_xxx.
 */
/*
int usb_stor_control_msg(struct us_data *us, unsigned int pipe,
			 u8 request, u8 requesttype, u16 value, u16 index, 
		 void *data, u16 size, int timeout)
*/
int usb_stor_control_msg(struct us_data *us, unsigned int pipe,
			 u8 request, u8 requesttype, u16 value, u16 index, 
		 void *data, u16 size, unsigned long timeout)
{
 int status;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_control_msg() Begin ..\n"));
 US_DEBUGP("%s: rq=%02x rqtype=%02x value=%04x index=%02x len=%u\n",
			__FUNCTION__, request, requesttype,
			value, index, size);

 #ifdef USBMS_BS_ASSERT
	if (size) {
		if ( ! (IS_VALID_ADDR(data)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_control_msg(1) : " ) ;
   		    iprintf("Error : Invalid Address !!\n" ) ;
	        iprintf("[ASSERT][USBMS] data addr = %08XH \n",
					data ) ; ;
	        iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	   } // if has length
 #endif // USBMS_BS_ASSERT

 #ifdef USBMS_BS_ASSERT
	if ( !(USB_IS_UN_CACHED(us)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_control_msg(2) : " ) ;
		    iprintf("Error : Not Un-Cached !!\n" ) ;
		    iprintf("[ASSERT][USBMS] us addr=%08XH \n", us ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
 #endif // USBMS_BS_ASSERT

 /* fill in the devrequest structure */
 us->cr->bRequestType = requesttype;
 us->cr->bRequest = request;
 us->cr->wValue = cpu_to_le16(value);
 us->cr->wIndex = cpu_to_le16(index);
 us->cr->wLength = cpu_to_le16(size);

 /* fill and submit the URB */
 usb_fill_control_urb(us->current_urb, us->pusb_dev, pipe, 
			 (unsigned char*) us->cr, data, size, 
			 usb_stor_blocking_completion, NULL);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_control_msg() : usb_stor_msg_common() ..\n"));
 status = usb_stor_msg_common(us, timeout);

 if(status == -ENODEV) { 
	USBMS_DBG_PLUG(iprintf(
		"[USBMS PLUG] usb_stor_control_msg() : during disconnect !! \n"
		)) ;
	// return -ENODEV ;
   }
 if ( status == -ENOMEM) { 
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] usb_stor_control_msg() : Error : Out of Memory !! \n"
		)) ;
	// return -ENOMEM ;
   }
 if ((status==-ENODEV) || (status==-ENOMEM)) { 
	return status ;
   }

 /* return the actual length of the data transferred if no error */
 if(status == 0) {
	status = us->current_urb->actual_length;
   }
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_control_msg() End.\n"));
 return status;
}

/* This is a version of usb_clear_halt() that allows early termination and
 * doesn't read the status from the device -- this is because some devices
 * crash their internal firmware when the status is requested after a halt.
 *
 * A definitive list of these 'bad' devices is too difficult to maintain or
 * make complete enough to be useful.  This problem was first observed on the
 * Hagiwara FlashGate DUAL unit.  However, bus traces reveal that neither
 * MacOS nor Windows checks the status after clearing a halt.
 *
 * Since many vendors in this space limit their testing to interoperability
 * with these two OSes, specification violations like this one are common.
 */
int usb_stor_clear_halt(struct us_data *us, unsigned int pipe)
{
	int result;
	int endp = usb_pipeendpoint(pipe);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_clear_halt() Begin ..\n"));
 #ifdef USBMS_BS_ASSERT
	if ( !(USB_IS_UN_CACHED(us)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_clear_halt() : " ) ;
		    iprintf("Error : Not Un-Cached !!\n" ) ;
		    iprintf("[ASSERT][USBMS] us addr=%08XH \n", us ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
 #endif // USBMS_BS_ASSERT

 if (usb_pipein (pipe))
		endp |= USB_DIR_IN;
 
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_clear_halt() : usb_stor_control_msg() ..\n"));
 result = usb_stor_control_msg(us, us->send_ctrl_pipe,
		USB_REQ_CLEAR_FEATURE, USB_RECIP_ENDPOINT, 0,
		/*
		endp, NULL, 0, 3*HZ);
		*/
		endp, NULL, 0, USBMS_CLEAR_TIMEOUT);
 

	/* reset the toggles and endpoint flags */
	usb_endpoint_running(us->pusb_dev, usb_pipeendpoint(pipe),
		usb_pipeout(pipe));
	usb_settoggle(us->pusb_dev, usb_pipeendpoint(pipe),
		usb_pipeout(pipe), 0);

	US_DEBUGP("%s: result = %d\n", __FUNCTION__, result);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_clear_halt() End.\n"));
	return result;
}


/*
 * Interpret the results of a URB transfer
 *
 * This function prints appropriate debugging messages, clears halts on
 * non-control endpoints, and translates the status to the corresponding
 * USB_STOR_XFER_xxx return code.
 */
static int interpret_urb_result(struct us_data *us, unsigned int pipe,
		unsigned int length, int result, unsigned int partial)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] interpret_urb_result() Begin .. \n")) ;
 /*
 US_DEBUGP("Status code %d; transferred %u/%u\n",
			result, partial, length);
 */
 USBMS_DBG_URB(iprintf(
	"[USBMS URB] Status code %d; transferred %u/%u\n",
			result, partial, length));

 switch (result) {

	/* no error code; did we send all the data? */
	case 0:
		if (partial != length) {
			US_DEBUGP("-- short transfer\n");
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] interpret_urb_result() End with USB_STOR_XFER_SHORT. \n")) ;
			return USB_STOR_XFER_SHORT;
		   }
		/*
		US_DEBUGP("-- transfer complete\n");
		*/
		USBMS_DBG_URB(iprintf("[USBMS URB] interpret_urb_result() : -- transfer complete \n")) ;
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] interpret_urb_result() End with USB_STOR_XFER_GOOD. \n")) ;
		return USB_STOR_XFER_GOOD;

	/* stalled */
	case -EPIPE:
		/* for control endpoints, (used by CB[I]) a stall indicates
		 * a failed command */
		if (usb_pipecontrol(pipe)) {
			US_DEBUGP("-- stall on control pipe\n");
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] interpret_urb_result() End with USB_STOR_XFER_STALLED. \n")) ;
			return USB_STOR_XFER_STALLED;
		}

		/* for other sorts of endpoint, clear the stall */
		US_DEBUGP("clearing endpoint halt for pipe 0x%x\n", pipe);
		if (usb_stor_clear_halt(us, pipe) < 0)
		   {
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] interpret_urb_result() End with USB_STOR_XFER_ERROR. \n")) ;
			return USB_STOR_XFER_ERROR;
		   }
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] interpret_urb_result() End with USB_STOR_XFER_STALLED. \n")) ;
		return USB_STOR_XFER_STALLED;

	/* timeout or excessively long NAK */
	case -ETIMEDOUT:
		US_DEBUGP("-- timeout or NAK\n");
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] interpret_urb_result() End with USB_STOR_XFER_ERROR. \n")) ;
		return USB_STOR_XFER_ERROR;

	/* babble - the device tried to send more than we wanted to read */
	case -EOVERFLOW:
		US_DEBUGP("-- babble\n");
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] interpret_urb_result() End with USB_STOR_XFER_LONG. \n")) ;
		return USB_STOR_XFER_LONG;

	/* the transfer was cancelled by abort, disconnect, or timeout */
	case -ECONNRESET:
		US_DEBUGP("-- transfer cancelled\n");
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] interpret_urb_result() End with USB_STOR_XFER_ERROR. \n")) ;
		return USB_STOR_XFER_ERROR;

	/* short scatter-gather read transfer */
	case -EREMOTEIO:
		US_DEBUGP("-- short read transfer\n");
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] interpret_urb_result() End with USB_STOR_XFER_SHORT. \n")) ;
		return USB_STOR_XFER_SHORT;

	/* abort or disconnect in progress */
	case -EIO:
		US_DEBUGP("-- abort or disconnect in progress\n");
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] interpret_urb_result() End with USB_STOR_XFER_ERROR. \n")) ;
		return USB_STOR_XFER_ERROR;

	/* the catch-all error case */
	default:
		US_DEBUGP("-- unknown error\n");
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] interpret_urb_result() End with USB_STOR_XFER_ERROR. \n")) ;
		return USB_STOR_XFER_ERROR;
	}
}

/*
 * Transfer one control message, without timeouts, but allowing early
 * termination.  Return codes are USB_STOR_XFER_xxx.
 */
int usb_stor_ctrl_transfer(struct us_data *us, unsigned int pipe,
		u8 request, u8 requesttype, u16 value, u16 index,
		void *data, u16 size)
{
 int result;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_ctrl_transfer() Begin ..\n"));

 #ifdef USBMS_BS_ASSERT
	if (size) {
		if ( ! (IS_VALID_ADDR(data)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_ctrl_transfer(1) : " ) ;
		    iprintf("Error : Invalid Address !!\n" ) ;
	        iprintf("[ASSERT][USBMS] data=%08XH \n",
					data ) ; ;
	        iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	   } // if has length
	if ( !(USB_IS_UN_CACHED(us)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_ctrl_transfer(2) : " ) ;
		    iprintf("Error : Not Un-Cached !!\n" ) ;
		    iprintf("[ASSERT][USBMS] us addr=%08XH \n", us ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
 #endif // USBMS_BS_ASSERT

 US_DEBUGP("[USBMS] %s: rq=%02x rqtype=%02x value=%04x index=%02x len=%u\n",
			__FUNCTION__, request, requesttype,
			value, index, size);

	/* fill in the devrequest structure */
	us->cr->bRequestType = requesttype;
	us->cr->bRequest = request;
	us->cr->wValue = cpu_to_le16(value);
	us->cr->wIndex = cpu_to_le16(index);
	us->cr->wLength = cpu_to_le16(size);

 /* fill and submit the URB */
 usb_fill_control_urb(us->current_urb, us->pusb_dev, pipe, 
			 (unsigned char*) us->cr, data, size, 
			 usb_stor_blocking_completion, NULL);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_ctrl_transfer() : usb_stor_msg_common() ..\n"));
 /*
 result = usb_stor_msg_common(us, 0);
 */
 result = usb_stor_msg_common(us, USBMS_TIMEOUT);

 if(result==-ENODEV) { 
	USBMS_DBG_PLUG(iprintf(
		"[USBMS PLUG] usb_stor_ctrl_transfer() : during disconnect !! \n"
		)) ;
	// return -ENODEV ;
   }
 if ( result == -ENOMEM) { 
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] usb_stor_ctrl_transfer() : Error : Out of Memory !! \n"
		)) ;
	// return -ENOMEM ;
   }
 if ((result==-ENODEV) || (result==-ENOMEM)) { 
	return result ;
   }

 /*
 status = usb_control_msg(usblp->usb_dev,
		dir ? usb_rcvctrlpipe(usblp->usb_dev, 0)
            : usb_sndctrlpipe(usblp->usb_dev, 0),
		// request, USB_TYPE_CLASS | dir | recip, value, usblp->ifnum,
        // buf, len, HZ * 5);
		request, type | dir | recip, value, usblp->ifnum,
        buf, len, USBLPT_CTLMSG_TIMEOUT);
         
 USBMS_DBG_INIT(iprintf(
	"[USBMS INIT] usb_stor_msg_common() : rq:%02XH dir:%d recip:%d "
	"value:%d len:%d result:%d \n",
	request, !!dir, recip, value, len, retval
    ));
 */

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_ctrl_transfer() End with Call interpret_urb_result() ..\n"));
 return interpret_urb_result(us, pipe, size, result,
			us->current_urb->actual_length);
}

/*
 * Receive one interrupt buffer, without timeouts, but allowing early
 * termination.  Return codes are USB_STOR_XFER_xxx.
 *
 * This routine always uses us->recv_intr_pipe as the pipe and
 * us->ep_bInterval as the interrupt interval.
 */
int usb_stor_intr_transfer(struct us_data *us, void *buf, unsigned int length)
{
 int result;
 unsigned int pipe = us->recv_intr_pipe;
 unsigned int maxp;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_intr_transfer() Begin ..\n"));
 US_DEBUGP("%s: xfer %u bytes\n", __FUNCTION__, length);

 #ifdef USBMS_BS_ASSERT
	if (length) {
		if ( ! (IS_VALID_ADDR(buf)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_intr_transfer(1) : " ) ;
   		    iprintf("Error : Invalid Address !!\n" ) ;
	        iprintf("[ASSERT][USBMS] buf=%08XH \n",
					buf ) ; ;
	        iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	   } // if has length
 #endif // USBMS_BS_ASSERT

 #ifdef USBMS_BS_ASSERT
	if ( !(USB_IS_UN_CACHED(us)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_intr_transfer(2) : " ) ;
		    iprintf("Error : Not Un-Cached !!\n" ) ;
		    iprintf("[ASSERT][USBMS] us addr=%08XH \n", us ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
 #endif // USBMS_BS_ASSERT

	/* calculate the max packet size */
	maxp = usb_maxpacket(us->pusb_dev, pipe, usb_pipeout(pipe));
	if (maxp > length)
		maxp = length;

 /* fill and submit the URB */
 usb_fill_int_urb(us->current_urb, us->pusb_dev, pipe, buf,
			maxp, usb_stor_blocking_completion, NULL,
			us->ep_bInterval);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_intr_transfer() : usb_stor_msg_common() ..\n"));
 /*
 result = usb_stor_msg_common(us, 0);
 */
 result = usb_stor_msg_common(us, USBMS_TIMEOUT);

 if(result==-ENODEV) { 
	USBMS_DBG_PLUG(iprintf(
		"[USBMS PLUG] usb_stor_intr_transfer() : during disconnect !! \n"
		)) ;
	// return -ENODEV ;
   }
 if ( result == -ENOMEM) { 
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] usb_stor_intr_transfer() : Error : Out of Memory !! \n"
		)) ;
	// return -ENOMEM ;
   }
 if ((result==-ENODEV) || (result==-ENOMEM)) { 
	return result ;
   }


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_intr_transfer() End with Call interpret_urb_result() ..\n"));
 return interpret_urb_result(us, pipe, length, result,
			us->current_urb->actual_length);
}

/*
 * Transfer one buffer via bulk pipe, without timeouts, but allowing early
 * termination.  Return codes are USB_STOR_XFER_xxx.  If the bulk pipe
 * stalls during the transfer, the halt is automatically cleared.
 */
int usb_stor_bulk_transfer_buf(struct us_data *us, unsigned int pipe,
	void *buf, unsigned int length, unsigned int *act_len)
{
 int result;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_bulk_transfer_buf() Begin .. \n")) ;
 /*
 US_DEBUGP("%s: xfer %u bytes\n", __FUNCTION__, length);
 */
 USBMS_DBG_URB(iprintf(
	"[USBMS URB] usb_stor_bulk_transfer_buf() : xfer %u bytes , buf=%08XH ..\n",
	length,buf));

 #ifdef USBMS_BS_ASSERT
	if (length) {
		if ( ! (IS_VALID_ADDR(buf)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_bulk_transfer_buf(1) : " ) ;
   		    iprintf("Error : Invalid Address !!\n" ) ;
	        iprintf("[ASSERT][USBMS] buf=%08XH \n",
					buf ) ; ;
	        iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	   } // if has length
 #endif // USBMS_BS_ASSERT

 #ifdef USBMS_BS_ASSERT
		if ( !(USB_IS_UN_CACHED(us)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_bulk_transfer_buf(2) : " ) ;
		    iprintf("Error : Not Un-Cached !!\n" ) ;
		    iprintf("[ASSERT][USBMS] us addr=%08XH \n", us ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
 #endif // USBMS_BS_ASSERT

 /* fill and submit the URB */
 usb_fill_bulk_urb(us->current_urb, us->pusb_dev, pipe, buf, length,
		      usb_stor_blocking_completion, NULL);
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_bulk_transfer_buf() : usb_stor_msg_common() ..\n"));
 /*
 result = usb_stor_msg_common(us, 0);
 */
 result = usb_stor_msg_common(us, USBMS_TIMEOUT);

 if(result==-ENODEV) { 
	USBMS_DBG_PLUG(iprintf(
		"[USBMS PLUG] usb_stor_bulk_transfer_buf() : during disconnect !! \n"
		)) ;
	// return -ENODEV ;
   }
 if ( result == -ENOMEM) { 
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] usb_stor_bulk_transfer_buf() : Error : Out of Memory !! \n"
		)) ;
	// return -ENOMEM ;
   }
 if ((result==-ENODEV) || (result==-ENOMEM)) { 
	return result ;
   }


 /* store the actual length of the data transferred */
 if (act_len)
		*act_len = us->current_urb->actual_length;

 USBMS_DBG_FLOW(iprintf(
	"[USBMS FLOW] usb_stor_bulk_transfer_buf() End with Call interpret_urb_result() .. \n"
	)) ;
 return interpret_urb_result(us, pipe, length, result, 
			us->current_urb->actual_length);
}

/*
 * Transfer a scatter-gather list via bulk transfer
 *
 * This function does basically the same thing as usb_stor_bulk_transfer_buf()
 * above, but it uses the usbcore scatter-gather library.
 */
/*
#ifndef NO_SCATTER_LIST
int usb_stor_bulk_transfer_sglist(struct us_data *us, unsigned int pipe,
		struct scatterlist *sg, 
	int num_sg, unsigned int length,
		unsigned int *act_len)
{
	int result;

	// don't submit s-g requests during abort/disconnect processing 
	if (us->flags & DONT_SUBMIT)
		return USB_STOR_XFER_ERROR;

	// initialize the scatter-gather request block
	US_DEBUGP("%s: xfer %u bytes, %d entries\n", __FUNCTION__,
			length, num_sg);
	#ifndef NO_SCATTER_LIST
		result = usb_sg_init(us->current_sg, us->pusb_dev, pipe, 0,
			#ifndef NO_SCATTER_LIST
				sg, 
			#endif // ndef NO_SCATTER_LIST
			num_sg, length, SLAB_NOIO);
		if (result) {
			US_DEBUGP("usb_sg_init returned %d\n", result);
			return USB_STOR_XFER_ERROR;
		}
	#endif // ndef NO_SCATTER_LIST

	// since the block has been initialized successfully, it's now
	// okay to cancel it 
	set_bit(US_FLIDX_SG_ACTIVE, &us->flags);

	// did an abort/disconnect occur during the submission? 
	if (us->flags & DONT_SUBMIT) {

		// cancel the request, if it hasn't been cancelled already 
		if (__test_and_clear_bit(US_FLIDX_SG_ACTIVE, &us->flags)) {
			US_DEBUGP("-- cancelling sg request\n");
			usb_sg_cancel(&us->current_sg);
		}
	}

	// wait for the completion of the transfer 
	usb_sg_wait(&us->current_sg);
	clear_bit(US_FLIDX_SG_ACTIVE, &us->flags);

	result = us->current_sg.status;
	if (act_len)
		*act_len = us->current_sg.bytes;
	return interpret_urb_result(us, pipe, length, result,
			us->current_sg.bytes);
}
#endif // ndef NO_SCATTER_LIST
*/
/*
 * Transfer an entire SCSI command's worth of data payload over the bulk
 * pipe.
 *
 * Note that this uses usb_stor_bulk_transfer_buf() and
 * usb_stor_bulk_transfer_sglist() to achieve its goals --
 * this function simply determines whether we're going to use
 * scatter-gather or not, and acts appropriately.
 */


int usb_stor_bulk_transfer_sg(struct us_data* us, unsigned int pipe,
		void *buf, unsigned int length_left,
 /*
 #ifndef NO_SCATTER_LIST
		int use_sg,
 #endif // ndef NO_SCATTER_LIST
 */
		int *residual)
{
 int result;
 unsigned int partial;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_bulk_transfer_sg() Begin .. \n")) ;

 /*
 #ifndef NO_SCATTER_LIST
	// are we scatter-gathering? 
	if (use_sg) {
		// use the usb core scatter-gather primitives 
		result = usb_stor_bulk_transfer_sglist(us, pipe,
				(struct scatterlist *) buf, use_sg,
				length_left, &partial);
		length_left -= partial;
	} else
 #endif // ndef NO_SCATTER_LIST
 */
	 {

	#ifdef USBMS_BS_ASSERT_CACHE
		if (buf) {
			if ( !(USB_IS_UN_CACHED(buf)) ) {
				iprintf("[ASSERT][USBMS] usb_stor_bulk_transfer_sg() : " ) ;
    		    iprintf("Error : Not Un-Cached !!\n" ) ;
		        iprintf("[ASSERT][USBMS] buf addr = %08XH \n", buf) ;
		        iprintf("[ASSERT][USBMS] Continue ..\n") ;
   			   }
		   } // if has data
	#endif // USBMS_BS_ASSERT

	// no scatter-gather, just make the request 
	USBMS_DBG_FLOW(iprintf(
		"[USBMS FLOW] usb_stor_bulk_transfer_sg() : usb_stor_bulk_transfer_buf() .. \n"
		)) ;

	result = usb_stor_bulk_transfer_buf(us, pipe, buf, 
										length_left, &partial);

	if(result==-ENODEV) { 
		USBMS_DBG_PLUG(iprintf(
			"[USBMS PLUG] usb_stor_bulk_transfer_sg(2) : during disconnect !! \n"
			)) ;
		// return -ENODEV ;
	   }
	if(result == -ENOMEM) { 
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_bulk_transfer_sg(2) : Error : Out of Memory !! \n"
			)) ;
		// return -ENOMEM ;
	   }
	if((result==-ENODEV) || (result==-ENOMEM)) { 
		return result ;
	   }

	length_left -= partial;
   }

 // store the residual and return the error code 
 if (residual)
		*residual = length_left;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_bulk_transfer_sg() End. \n")) ;
 return result;
}

/***********************************************************************
 * Transport routines
 ***********************************************************************/

/* Invoke the transport and basic error-handling/recovery methods
 *
 * This is used by the protocol layers to actually send the message to
 * the device and receive the response.
 */

// return : 1 : ok , 0 : error , -ENODEV : un-plug

int usb_stor_invoke_transport(Scsi_Cmnd *srb, struct us_data *us)
{
 int need_auto_sense;
 int result;
 // unsigned char *tmp_sense_buffer_pr ; // B. S. Liaw add this !!


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_invoke_transport() Begin .. \n"));

 // 2004/1/12 B. S. Liaw
 if ( !us || (us->flags & DONT_SUBMIT)) {
	USBMS_DBG_INFO(iprintf("[USBMS INFO] usb_stor_invoke_transport(1) : DONT_SUBMIT \n" )) ;
	USBMS_DBG_PLUG(iprintf("[USBMS PLUG] usb_stor_invoke_transport(2) : during disconnect !! \n" )) ;
	return -ENODEV ;
   }

 /* send the command to the transport layer */
 srb->resid = 0;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_invoke_transport() : us->transport() .. \n"));
 result = us->transport(srb, us);
 if(result==-ENODEV) { 
	USBMS_DBG_PLUG(iprintf(
		"[USBMS PLUG] usb_stor_invoke_transport(3) : during disconnect !! \n"
		)) ;
	// return -ENODEV ;
   }
 if ( result == -ENOMEM) { 
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] usb_stor_invoke_transport(2) : Error : Out of Memory !! \n"
		)) ;
	// return -ENOMEM ;
   }
 if ((result==-ENODEV) || (result==-ENOMEM)) { 
	return result ;
   }

 /* if the command gets aborted by the higher layers, we need to
  * short-circuit all other processing
  */
 if (us->sm_state == US_STATE_ABORTING) {
	// US_DEBUGP("-- command was aborted\n");
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] usb_stor_invoke_transport(3) : Error : command US_STATE_ABORTING !! \n"
		)) ;
	goto Handle_Abort;
   }

 /* if there is a transport error, reset and don't auto-sense */
 if(result == USB_STOR_TRANSPORT_ERROR) {
	// US_DEBUGP("-- transport indicates error, resetting\n");
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] usb_stor_invoke_transport(4) : Error : USB_STOR_TRANSPORT_ERROR !! \n"
		)) ;

	// 2004/01/12 B. S. Liaw
	if (!us) {
		/*
		US_DEBUGP("No command during disconnect\n");
		*/
		USBMS_DBG_PLUG(iprintf(
			"[USBMS PLUG] usb_stor_invoke_transport(6) : No command during disconnect \n"
			)) ;
		// srb->result = DID_BAD_TARGET << 16;
		return -ENODEV ;
	   }
	if (test_bit(US_FLIDX_DISCONNECTING, &us->flags)) {
		/*
		US_DEBUGP("No command during disconnect\n");
		*/
		USBMS_DBG_PLUG(iprintf(
			"[USBMS PLUG] usb_stor_invoke_transport(7) : No command during disconnect \n"
			)) ;
		// srb->result = DID_BAD_TARGET << 16;
		return -ENODEV ;
	   }

	if(us && us->transport_reset ) {
		dprintf("[USBMS] Error : Transport Error , Reset .. \n") ;
		us->transport_reset(us);
	   }
	  else {
		USBMS_DBG_PLUG(iprintf(
			"[USBMS PLUG] usb_stor_invoke_transport(8) : No Reset , during disconnect ??\n"
			)) ;
	   }
	srb->result = DID_ERROR << 16;
	return 1 ;
   }

 /* if the transport provided its own sense data, don't auto-sense */
 if(result == USB_STOR_TRANSPORT_NO_SENSE) {
	srb->result = SAM_STAT_CHECK_CONDITION;
	return 1 ;
   }

 /* Determine if we need to auto-sense
  *
  * I normally don't use a flag like this, but it's almost impossible
  * to understand what's going on here if I don't.
  */
 need_auto_sense = 0;

 /*
  * If we're running the CB transport, which is incapable
  * of determining status on it's own, we need to auto-sense almost
  * every time.
  */
 if(us->protocol == US_PR_CB || us->protocol == US_PR_DPCM_USB) {
	US_DEBUGP("-- CB transport device requiring auto-sense\n");
	need_auto_sense = 1;

	/* There are some exceptions to this.  Notably, if this is
	 * a UFI device and the command is REQUEST_SENSE or INQUIRY,
	 * then it is impossible to truly determine status.
	 */
	if (us->subclass == US_SC_UFI &&
	    ((srb->cmnd[0] == REQUEST_SENSE) ||
	     (srb->cmnd[0] == INQUIRY))) {
		US_DEBUGP("** no auto-sense for a special command\n");
		need_auto_sense = 0;
	   }
   }

 /*
  * If we have a failure, we're going to do a REQUEST_SENSE 
  * automatically.  Note that we differentiate between a command
  * "failure" and an "error" in the transport mechanism.
  */
 if (result == USB_STOR_TRANSPORT_FAILED) {
	US_DEBUGP("-- transport indicates command failure\n");
	need_auto_sense = 1;
   }

 /*
  * Also, if we have a short transfer on a command that can't have
  * a short transfer, we're going to do this.
  */
 if ((srb->resid > 0) &&
    !((srb->cmnd[0] == REQUEST_SENSE) ||
      (srb->cmnd[0] == INQUIRY) ||
      (srb->cmnd[0] == MODE_SENSE) ||
      (srb->cmnd[0] == LOG_SENSE) ||
      (srb->cmnd[0] == MODE_SENSE_10))) {
	US_DEBUGP("-- unexpectedly short transfer\n");
	need_auto_sense = 1;
   }

 /* Now, if we need to do the auto-sense, let's do it */
 if(need_auto_sense) {
	int temp_result;
	void* old_request_buffer;
	/*
	#ifndef NO_SCATTER_LIST
		unsigned short old_sg;
	#endif // ndef NO_SCATTER_LIST
	*/
	unsigned old_request_bufflen;
	unsigned char old_sc_data_direction;
	unsigned char old_cmd_len;
	unsigned char old_cmnd[MAX_COMMAND_SIZE];
	unsigned long old_serial_number;

	US_DEBUGP("[USBMS] Issuing auto-REQUEST_SENSE\n");

	/* save the old command */
	memcpy(old_cmnd, srb->cmnd, MAX_COMMAND_SIZE);
	old_cmd_len = srb->cmd_len;

	/* set the command and the LUN */
	
	// scsi_cmnd.h : srb->cmnd : unsigned char cmnd[MAX_COMMAND_SIZE];
	//  tmp_sense_buffer_pr = srb->cmd->sense_buffer ; // B. S. Liaw : backup
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_invoke_transport() : memset() .. \n" )) ;
	memset(srb->cmnd, 0, MAX_COMMAND_SIZE);
	// srb->cmd->sense_buffer =	tmp_sense_buffer_pr ; // B. S. Liaw : restore
	srb->cmnd[0] = REQUEST_SENSE;
	srb->cmnd[1] = old_cmnd[1] & 0xE0;
	srb->cmnd[4] = 18;

	/* FIXME: we must do the protocol translation here */
	if (us->subclass == US_SC_RBC || us->subclass == US_SC_SCSI)
			srb->cmd_len = 6;
	  else
			srb->cmd_len = 12;

	/* set the transfer direction */
	old_sc_data_direction = srb->sc_data_direction;
	srb->sc_data_direction = SCSI_DATA_READ;

	/* use the new buffer we have */
	old_request_buffer = srb->request_buffer;
	// B. S. Liaw : request_buffer need uncached
	// B. S. Liaw : sense_buffer need uncached
	srb->request_buffer = srb->sense_buffer;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] usb_stor_invoke_transport() : Set Value : request_buffer=%08XH .. \n",
		srb->request_buffer
		));

	/* set the buffer length for transfer */
	old_request_bufflen = srb->request_bufflen;
	srb->request_bufflen = 18;

	/* set up for no scatter-gather use */
	/*
	#ifndef NO_SCATTER_LIST
		old_sg = srb->use_sg;
		srb->use_sg = 0;
	#endif // ndef NO_SCATTER_LIST
	*/
	/* change the serial number -- toggle the high bit*/
	old_serial_number = srb->serial_number;
	srb->serial_number ^= 0x80000000;

	/* issue the auto-sense command */
	USBMS_DBG_FLOW(iprintf(
		"[USBMS FLOW] usb_stor_invoke_transport() : us->transport() .. \n"
		));
	temp_result = us->transport(us->srb, us);
	if ( temp_result == -ENODEV) { 
		USBMS_DBG_PLUG(iprintf(
			"[USBMS PLUG] usb_stor_invoke_transport(9) : during disconnect !! \n"
			)) ;
		return -ENODEV ;
	   }

	/* let's clean up right away */
	srb->request_buffer = old_request_buffer;
	USBMS_DBG_MEM(iprintf(
		"[USBMS MEM] usb_stor_invoke_transport() : Set Value : request_buffer=%08XH .. \n",
		srb->request_buffer
		));
	srb->request_bufflen = old_request_bufflen;
	/*
	#ifndef NO_SCATTER_LIST
		srb->use_sg = old_sg;
	#endif // ndef NO_SCATTER_LIST
	*/
	srb->serial_number = old_serial_number;
	srb->sc_data_direction = old_sc_data_direction;
	srb->cmd_len = old_cmd_len;
	memcpy(srb->cmnd, old_cmnd, MAX_COMMAND_SIZE);

	if (us->sm_state == US_STATE_ABORTING) {
		/*
		US_DEBUGP("-- auto-sense aborted\n");
		*/
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_invoke_transport(5) : Error : auto-sense US_STATE_ABORTING !! \n"
			)) ;
		goto Handle_Abort;
	   }
	if (temp_result != USB_STOR_TRANSPORT_GOOD) {
		US_DEBUGP("-- auto-sense failure\n");

		/* we skip the reset if this happens to be a
		 * multi-target device, since failure of an
		 * auto-sense is perfectly valid
		 */
		if (!(us->flags & US_FL_SCM_MULT_TARG)) {
			dprintf("[USBMS] Error : Auto-Sense Failure , Reset .. \n") ;
			us->transport_reset(us);
		   }
		srb->result = DID_ERROR << 16;
		return 1 ;
	   }

	US_DEBUGP("-- Result from auto-sense is %d\n", temp_result);
	US_DEBUGP("-- code: 0x%x, key: 0x%x, ASC: 0x%x, ASCQ: 0x%x\n",
			  srb->sense_buffer[0],
			  srb->sense_buffer[2] & 0xf,
			  srb->sense_buffer[12], 
			  srb->sense_buffer[13]);
	#ifdef CONFIG_USB_STORAGE_DEBUG
			usb_stor_show_sense(
			  srb->sense_buffer[2] & 0xf,
			  srb->sense_buffer[12], 
			  srb->sense_buffer[13]);
	#endif

	/* set the result so the higher layers expect this data */
	srb->result = SAM_STAT_CHECK_CONDITION;

	/* If things are really okay, then let's show that */
	if ((srb->sense_buffer[2] & 0xf) == 0x0)
		srb->result = SAM_STAT_GOOD;
   } else /* if (need_auto_sense) */
	srb->result = SAM_STAT_GOOD;

 /* Regardless of auto-sense, if we _know_ we have an error
  * condition, show that in the result code
  */
 if (result == USB_STOR_TRANSPORT_FAILED)
		srb->result = SAM_STAT_CHECK_CONDITION;

 /* If we think we're good, then make sure the sense data shows it.
  * This is necessary because the auto-sense for some devices always
  * sets byte 0 == 0x70, even if there is no error
  */
 if ((us->protocol == US_PR_CB || us->protocol == US_PR_DPCM_USB) && 
	    (result == USB_STOR_TRANSPORT_GOOD) &&
	    ((srb->sense_buffer[2] & 0xf) == 0x0))
		srb->sense_buffer[0] = 0x0;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_invoke_transport() End. \n"));
 return 1 ;

 /* abort processing: the bulk-only transport requires a reset
  * following an abort */

 // -------------------
 Handle_Abort:
 // --------------------

 srb->result = DID_ABORT << 16;
 if (us->protocol == US_PR_BULK) {

		/* permit the reset transfer to take place */
		clear_bit(US_FLIDX_ABORTING, &us->flags);
		dprintf("[USBMS] Error : Transport Abort , Reset .. \n") ;
		us->transport_reset(us);
	}
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_invoke_transport() End with Failed !! \n"));
 return 1 ;
}

/* Stop the current URB transfer */
void usb_stor_stop_transport(struct us_data *us)
{
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] usb_stor_stop_transport() Begin ..\n" )) ;
 // US_DEBUGP("%s called\n", __FUNCTION__);

 /* If the state machine is blocked waiting for an URB,
  * let's wake it up.  The test_and_clear_bit() call
  * guarantees that if a URB has just been submitted,
  * it won't be cancelled more than once. */
 if (__test_and_clear_bit(US_FLIDX_URB_ACTIVE, &us->flags)) {
	// 	US_DEBUGP("-- cancelling URB\n");
	USBMS_DBG_PLUG(iprintf("[USBMS PLUG] usb_stor_stop_transport() -- cancelling URB .. \n" )) ;
	usb_unlink_urb(us->current_urb);
   }

	/* If we are waiting for a scatter-gather operation, cancel it. */
 if (__test_and_clear_bit(US_FLIDX_SG_ACTIVE, &us->flags)) {
	//	US_DEBUGP("-- cancelling sg request\n");
	USBMS_DBG_PLUG(iprintf("[USBMS PLUG] usb_stor_stop_transport() -- cancelling sg request .. \n" )) ;
	usb_sg_cancel(&us->current_sg);
   }
 USBMS_DBG_PLUG(iprintf("[USBMS PLUG] usb_stor_stop_transport() End.\n" )) ;
}


/*
 * Control/Bulk/Interrupt transport
 */

int usb_stor_CBI_transport(Scsi_Cmnd *srb, struct us_data *us)
{
 unsigned int transfer_length = srb->request_bufflen;
 int result;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_CBI_transport() Begin .. \n")) ;
 /* COMMAND STAGE */
 /* let's send the command via the control pipe */
 result = usb_stor_ctrl_transfer(us, us->send_ctrl_pipe,
				      US_CBI_ADSC, 
				      USB_TYPE_CLASS | USB_RECIP_INTERFACE, 0, 
				      us->ifnum, srb->cmnd, srb->cmd_len);

	/* check the return code for the command */
	US_DEBUGP("Call to usb_stor_ctrl_transfer() returned %d\n", result);

	/* if we stalled the command, it means command failed */
	if (result == USB_STOR_XFER_STALLED) {
		return USB_STOR_TRANSPORT_FAILED;
	}

	/* Uh oh... serious problem here */
	if (result != USB_STOR_XFER_GOOD) {
		return USB_STOR_TRANSPORT_ERROR;
	}

	/* DATA STAGE */
	/* transfer the data payload for this command, if one exists*/
	if (transfer_length) {
		unsigned int pipe = srb->sc_data_direction == SCSI_DATA_READ ? 
				us->recv_bulk_pipe : us->send_bulk_pipe;

	USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_CBI_transport() .. \n"
			)) ;
 #ifdef USBMS_BS_ASSERT
	if (transfer_length) {
		if ( ! (IS_VALID_ADDR(srb->request_buffer)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_CBI_transport(1) : " ) ;
   		    iprintf("Error : Invalid Address !!\n" ) ;
	        iprintf("[ASSERT][USBMS] srb->request_buffer=%08XH \n",
					srb->request_buffer ) ; ;
	        iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	   } // if has length
 #endif // USBMS_BS_ASSERT

 #ifdef USBMS_BS_ASSERT
		if ( !(USB_IS_UN_CACHED(us)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_CBI_transport(2) : " ) ;
		    iprintf("Error : Not Un-Cached !!\n" ) ;
		    iprintf("[ASSERT][USBMS] us=%08XH \n", us ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
 #endif // USBMS_BS_ASSERT

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_CBI_transport() : usb_stor_bulk_transfer_sg() .. \n")) ;
		result = usb_stor_bulk_transfer_sg(us, pipe,
					srb->request_buffer, transfer_length,
					/*
					#ifndef NO_SCATTER_LIST
						srb->use_sg,
					#endif // ndef NO_SCATTER_LIST
					*/
					&srb->resid);

		US_DEBUGP("CBI data stage result is 0x%x\n", result);
		if( result == -ENODEV) {
			USBMS_DBG_PLUG(iprintf(
				"[USBMS PLUG] usb_stor_CBI_transport() : during disconnect !! \n"
				)) ;
			return -ENODEV ;
		   }
		if (result > USB_STOR_XFER_STALLED)
			return USB_STOR_TRANSPORT_ERROR;
	}

	/* STATUS STAGE */
	result = usb_stor_intr_transfer(us, us->iobuf, 2);
	US_DEBUGP("Got interrupt data (0x%x, 0x%x)\n", 
			us->iobuf[0], us->iobuf[1]);
	if (result != USB_STOR_XFER_GOOD)
		return USB_STOR_TRANSPORT_ERROR;

	/* UFI gives us ASC and ASCQ, like a request sense
	 *
	 * REQUEST_SENSE and INQUIRY don't affect the sense data on UFI
	 * devices, so we ignore the information for those commands.  Note
	 * that this means we could be ignoring a real error on these
	 * commands, but that can't be helped.
	 */
	if (us->subclass == US_SC_UFI) {
		if (srb->cmnd[0] == REQUEST_SENSE ||
		    srb->cmnd[0] == INQUIRY)
			return USB_STOR_TRANSPORT_GOOD;
		else {
			if (us->iobuf[0])
				return USB_STOR_TRANSPORT_FAILED;
			else
				return USB_STOR_TRANSPORT_GOOD;
		}
	}

	/* If not UFI, we interpret the data as a result code 
	 * The first byte should always be a 0x0
	 * The second byte & 0x0F should be 0x0 for good, otherwise error 
	 */
	if (us->iobuf[0]) {
		US_DEBUGP("CBI IRQ data showed reserved bType %d\n",
				us->iobuf[0]);
		return USB_STOR_TRANSPORT_ERROR;
	}

	switch (us->iobuf[1] & 0x0F) {
		case 0x00: 
			return USB_STOR_TRANSPORT_GOOD;
		case 0x01: 
			return USB_STOR_TRANSPORT_FAILED;
		default: 
			return USB_STOR_TRANSPORT_ERROR;
	}

	/* we should never get here, but if we do, we're in trouble */
	return USB_STOR_TRANSPORT_ERROR;
}

/*
 * Control/Bulk transport
 */
int usb_stor_CB_transport(Scsi_Cmnd *srb, struct us_data *us)
{
 unsigned int transfer_length = srb->request_bufflen;
 int result;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_CB_transport() Begin .. \n")) ;
 /* COMMAND STAGE */
 /* let's send the command via the control pipe */
 result = usb_stor_ctrl_transfer(us, us->send_ctrl_pipe,
				      US_CBI_ADSC, 
				      USB_TYPE_CLASS | USB_RECIP_INTERFACE, 0, 
				      us->ifnum, srb->cmnd, srb->cmd_len);

 /* check the return code for the command */
 US_DEBUGP("Call to usb_stor_ctrl_transfer() returned %d\n", result);

 /* if we stalled the command, it means command failed */
 if (result == USB_STOR_XFER_STALLED) {
		return USB_STOR_TRANSPORT_FAILED;
	}

 /* Uh oh... serious problem here */
 if (result != USB_STOR_XFER_GOOD) {
		return USB_STOR_TRANSPORT_ERROR;
	}

 /* DATA STAGE */
 /* transfer the data payload for this command, if one exists*/
 if (transfer_length) {
	unsigned int pipe = srb->sc_data_direction == SCSI_DATA_READ ? 
				us->recv_bulk_pipe : us->send_bulk_pipe;

	USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_CB_transport() .. \n"
			)) ;
	#ifdef USBMS_BS_ASSERT
		if (transfer_length) {
			if ( ! (IS_VALID_ADDR(srb->request_buffer)) ) {
				iprintf("[ASSERT][USBMS] usb_stor_CB_transport(1) : " ) ;
   		    	iprintf("Error : Invalid Address !!\n" ) ;
		        iprintf("[ASSERT][USBMS] srb->request_buffer=%08XH \n",
						srb->request_buffer ) ; ;
	    	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
			   }
		   } // if has length
	#endif // USBMS_BS_ASSERT

	#ifdef USBMS_BS_ASSERT
		if ( !(USB_IS_UN_CACHED(us)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_CB_transport(2) : " ) ;
		    iprintf("Error : Not Un-Cached !!\n" ) ;
		    iprintf("[ASSERT][USBMS] us addr=%08XH \n", us ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	#endif // USBMS_BS_ASSERT

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_CB_transport() : usb_stor_bulk_transfer_sg() .. \n")) ;
	result = usb_stor_bulk_transfer_sg(us, pipe,
					srb->request_buffer, transfer_length,
				/*
				#ifndef NO_SCATTER_LIST
					srb->use_sg,
				#endif // ndef NO_SCATTER_LIST
				*/
				&srb->resid);
	US_DEBUGP("CB data stage result is 0x%x\n", result);
	if( result == -ENODEV) {
		USBMS_DBG_PLUG(iprintf(
			"[USBMS PLUG] usb_stor_CB_transport() : during disconnect !! \n"
			)) ;
		return -ENODEV ;
	   }

	if (result > USB_STOR_XFER_STALLED)
			return USB_STOR_TRANSPORT_ERROR;
   }

 /* STATUS STAGE */
 /* NOTE: CB does not have a status stage.  Silly, I know.  So
  * we have to catch this at a higher level.
  */
 return USB_STOR_TRANSPORT_GOOD;
}

/*
 * Bulk only transport
 */

/* Determine what the maximum LUN supported is */
#ifndef USBMS_NO_PROBE_MAXLUN
int usb_stor_Bulk_max_lun(struct us_data *us)
{
 int result;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_max_lun() Begin .. \n"));

 #ifdef USBMS_BS_ASSERT
	if ( !(USB_IS_UN_CACHED(us)) ) {
		iprintf("[ASSERT][USBMS] usb_stor_Bulk_max_lun() : " ) ;
	    iprintf("Error : Not Un-Cached !!\n" ) ;
	    iprintf("[ASSERT][USBMS] us addr=%08XH \n", us ) ;
	    iprintf("[ASSERT][USBMS] Continue ..\n") ;
	   }
 #endif // USBMS_BS_ASSERT

 /* issue the command */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_max_lun() : usb_stor_control_msg() ..\n"));
 result = usb_stor_control_msg(us, us->recv_ctrl_pipe,
				 US_BULK_GET_MAX_LUN, 
				 USB_DIR_IN | USB_TYPE_CLASS | 
				 USB_RECIP_INTERFACE,
				/*
				 0, us->ifnum, us->iobuf, 1, HZ);
				*/
				 0, us->ifnum, us->iobuf, 1, USBMS_MAXLUN_TIMEOUT);

 /*
 US_DEBUGP("GetMaxLUN command result is %d, data is %d\n", 
		  result, us->iobuf[0]);
 */
 USBMS_DBG_URB(iprintf(
	"[USBMS URB] usb_stor_Bulk_max_lun() : GetMaxLUN result %d, data %d \n",
	result, us->iobuf[0]
	)) ;

 /* if we have a successful request, return the result */
 if(result == 1) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_max_lun() End. \n"));
	return us->iobuf[0];
   }

 /* return the default -- no LUNs */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_max_lun() End with no LUNs. \n"));
 return 0;
}
#endif // ndef USBMS_NO_PROBE_MAXLUN

int usb_stor_Bulk_transport(Scsi_Cmnd *srb, struct us_data *us)
{
 struct bulk_cb_wrap *bcb = (struct bulk_cb_wrap *) us->iobuf;
 struct bulk_cs_wrap *bcs = (struct bulk_cs_wrap *) us->iobuf;
 unsigned int transfer_length = srb->request_bufflen;
 int result;
 int fake_sense = 0;


 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_transport() Begin .. \n")) ;

 #ifdef USBMS_DBG_READ_ONLY
	#warning: USBMS_DBG_READ_ONLY : ON !!
	if (srb->sc_data_direction == SCSI_DATA_WRITE) {
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_Bulk_transport(1) : Error : Read Only , Return USB_STOR_TRANSPORT_GOOD !! \n"
			)) ;
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_transport(1) End. \n")) ;
		return USB_STOR_TRANSPORT_GOOD;
	   }
 #endif // USBMS_DBG_READ_ONLY

 /* set up the command wrapper */
 bcb->Signature = cpu_to_le32(US_BULK_CB_SIGN);
 bcb->DataTransferLength = cpu_to_le32(transfer_length);
 bcb->Flags = srb->sc_data_direction == SCSI_DATA_READ ? 1 << 7 : 0;
 bcb->Tag = srb->serial_number;
 bcb->Lun = srb->device->lun;
 if (us->flags & US_FL_SCM_MULT_TARG)
		bcb->Lun |= srb->device->id << 4;
 bcb->Length = srb->cmd_len;

 /* copy the command payload */
 memset(bcb->CDB, 0, sizeof(bcb->CDB));
 memcpy(bcb->CDB, srb->cmnd, bcb->Length);

#if 0
 iprintf("DataTransferLength=%lu, Flags=%02XH, Tag=%lu\n", bcb->DataTransferLength, bcb->Flags, bcb->Tag);
 iprintf("srb->cmnd(%d): ", srb->cmd_len);
 for (result=0; result < srb->cmd_len; result++)
 {
	iprintf("%02X ", bcb->CDB[result]);
 }
 iprintf("\n");
#endif

 /* send it to out endpoint */
 /*
 US_DEBUGP("Bulk Command S 0x%x T 0x%x L %d F %d Trg %d LUN %d CL %d\n",
 */
 USBMS_DBG_URB(iprintf(
	"[USBMS URB] Bulk Command S 0x%x T 0x%x L %d F %d Trg %d LUN %d CL %d\n",
	le32_to_cpu(bcb->Signature), bcb->Tag,
	le32_to_cpu(bcb->DataTransferLength), bcb->Flags,
	(bcb->Lun >> 4), (bcb->Lun & 0x0F), 
	/*
	bcb->Length);
	*/
	bcb->Length));


 USBMS_DBG_URB(iprintf("[USBMS URB] usb_stor_Bulk_transport() : --1-- Command ..\n")) ;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_transport(3) : usb_stor_bulk_transfer_buf() .. \n")) ;
 result = usb_stor_bulk_transfer_buf(us, us->send_bulk_pipe,
				bcb, US_BULK_CB_WRAP_LEN, NULL);
 US_DEBUGP("Bulk command transfer result=%d\n", result);

 if(result==-ENODEV) { 
	USBMS_DBG_PLUG(iprintf(
		"[USBMS PLUG] usb_stor_Bulk_transport(2) : during disconnect !! \n"
		)) ;
	// return -ENODEV ;
   }
 if(result == -ENOMEM) { 
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] usb_stor_Bulk_transport(2) : Error : Out of Memory !! \n"
		)) ;
	// return -ENOMEM ;
   }
 if((result==-ENODEV) || (result==-ENOMEM)) { 
	return result ;
   }


 if(result != USB_STOR_XFER_GOOD) {
		return USB_STOR_TRANSPORT_ERROR;
   }
 /* DATA STAGE */
 /* send/receive data payload, if there is any */
 if (transfer_length) {
	unsigned int pipe = srb->sc_data_direction == SCSI_DATA_READ ? 
				us->recv_bulk_pipe : us->send_bulk_pipe;

	USBMS_DBG_URB(iprintf("[USBMS URB] usb_stor_Bulk_transport() : --2-- Data ..\n")) ;
	// USBMS_DBG_INIT(iprintf(
	//		"[USBMS INIT] usb_stor_Bulk_transport() .. \n"
	//		)) ;

	#ifdef USBMS_BS_ASSERT
		if ( ! (IS_VALID_ADDR(srb->request_buffer)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_Bulk_transport(1) : " ) ;
   		    iprintf("Error : Invalid Address !!\n" ) ;
	        iprintf("[ASSERT][USBMS] srb->request_buffer=%08XH \n",
						srb->request_buffer ) ; ;
	        iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }

		if ( !(USB_IS_UN_CACHED(us)) ) {
			iprintf("[ASSERT][USBMS] usb_stor_Bulk_transport(2) : " ) ;
		    iprintf("Error : Not Un-Cached !!\n" ) ;
		    iprintf("[ASSERT][USBMS] us addr=%08XH \n", us ) ;
		    iprintf("[ASSERT][USBMS] Continue ..\n") ;
		   }
	#endif // USBMS_BS_ASSERT

	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_transport(4) : usb_stor_bulk_transfer_sg() .. \n")) ;
	result = usb_stor_bulk_transfer_sg(us, pipe,
					srb->request_buffer, transfer_length,
				/*
				#ifndef NO_SCATTER_LIST
					srb->use_sg, 
				#endif // ndef NO_SCATTER_LIST
				*/
				&srb->resid);

	US_DEBUGP("Bulk data transfer result 0x%x\n", result);
	if( result == -ENODEV) {
		USBMS_DBG_PLUG(iprintf(
			"[USBMS PLUG] usb_stor_Bulk_transport(3) : during disconnect !! \n"
			)) ;
		return -ENODEV ;
	   }

	if (result == USB_STOR_XFER_ERROR)
	   {
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_transport(5) End with Fail !! \n")) ;
		return USB_STOR_TRANSPORT_ERROR;
	   }

		/* If the device tried to send back more data than the
		 * amount requested, the spec requires us to transfer
		 * the CSW anyway.  Since there's no point retrying the
		 * the command, we'll return fake sense data indicating
		 * Illegal Request, Invalid Field in CDB.
		 */
	if (result == USB_STOR_XFER_LONG)
		fake_sense = 1;
   }
  else {
	USBMS_DBG_URB(iprintf("[USBMS URB] usb_stor_Bulk_transport() : --2-- No Data !! \n")) ;
   }

 /* See flow chart on pg 15 of the Bulk Only Transport spec for
  * an explanation of how this code works.
  */

 /* get CSW for device status */
 USBMS_DBG_URB(iprintf("[USBMS URB] usb_stor_Bulk_transport() : --3-- get CSW ..\n")) ;
 /*
 US_DEBUGP("Attempting to get CSW...\n");
 */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_transport(6) : usb_stor_bulk_transfer_buf() .. \n")) ;

 result = usb_stor_bulk_transfer_buf(us, us->recv_bulk_pipe,
				bcs, US_BULK_CS_WRAP_LEN, NULL);

 if(result==-ENODEV) { 
	USBMS_DBG_PLUG(iprintf(
		"[USBMS PLUG] usb_stor_Bulk_transport(4) : during disconnect !! \n"
		)) ;
	// return -ENODEV ;
   }
 if(result == -ENOMEM) { 
	USBMS_DBG_INFO(iprintf(
		"[USBMS INFO] usb_stor_Bulk_transport(3) : Error : Out of Memory !! \n"
		)) ;
	// return -ENOMEM ;
   }
 if((result==-ENODEV) || (result==-ENOMEM)) { 
	return result ;
   }

 /* did the attempt to read the CSW fail? */
 if(result == USB_STOR_XFER_STALLED) {
	/* get the status again */
	USBMS_DBG_URB(iprintf("[USBMS URB] usb_stor_Bulk_transport() : --3-- get CSW (2nd try) .. \n")) ;
	/*
	US_DEBUGP("Attempting to get CSW (2nd try)...\n");
	*/
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_transport(7) : usb_stor_bulk_transfer_buf() .. \n")) ;
	result = usb_stor_bulk_transfer_buf(us, us->recv_bulk_pipe,
				bcs, US_BULK_CS_WRAP_LEN, NULL);

	if(result==-ENODEV) { 
		USBMS_DBG_PLUG(iprintf(
			"[USBMS PLUG] usb_stor_Bulk_transport(5) : during disconnect !! \n"
			)) ;
		// return -ENODEV ;
	   }
	if (result == -ENOMEM) { 
		USBMS_DBG_INFO(iprintf(
			"[USBMS INFO] usb_stor_Bulk_transport(4) : Error : Out of Memory !! \n"
			)) ;
		// return -ENOMEM ;
	   }
	if ((result==-ENODEV) || (result==-ENOMEM)) { 
		return result ;
	   }
   }

 /* if we still have a failure at this point, we're in trouble */
 US_DEBUGP("Bulk status result = %d\n", result);
 if (result != USB_STOR_XFER_GOOD) {
	USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_transport(8) End with Fail !! \n")) ;
	return USB_STOR_TRANSPORT_ERROR;
   }

 /* check bulk status */
 USBMS_DBG_URB(iprintf("[USBMS URB] usb_stor_Bulk_transport() : --4-- check bulk status ..\n")) ;
 US_DEBUGP("Bulk Status S 0x%x T 0x%x R %d Stat 0x%x\n",
			le32_to_cpu(bcs->Signature), bcs->Tag, 
			bcs->Residue, bcs->Status);
 if ((bcs->Signature != cpu_to_le32(US_BULK_CS_SIGN) &&
		    bcs->Signature != cpu_to_le32(US_BULK_CS_OLYMPUS_SIGN)) ||
			bcs->Tag != srb->serial_number || 
			bcs->Status > US_BULK_STAT_PHASE) {
		US_DEBUGP("Bulk logical error\n");
		USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_transport(9) End with Fail !! \n")) ;
		return USB_STOR_TRANSPORT_ERROR;
	}

 /* based on the status code, we report good or bad */
 switch (bcs->Status) {
	case US_BULK_STAT_OK:
			/* device babbled -- return fake sense data */
			if (fake_sense) {
				memcpy(srb->sense_buffer, 
				       usb_stor_sense_invalidCDB, 
				       sizeof(usb_stor_sense_invalidCDB));
				USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_transport(10) End with Fail !! \n")) ;
				return USB_STOR_TRANSPORT_NO_SENSE;
			}

			/* command good -- note that data could be short */
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_transport() End. \n")) ;
			return USB_STOR_TRANSPORT_GOOD;

	case US_BULK_STAT_FAIL:
			/* command failed */
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_transport(11) End with Fail !! \n")) ;
			return USB_STOR_TRANSPORT_FAILED;

	case US_BULK_STAT_PHASE:
			/* phase error -- note that a transport reset will be
			 * invoked by the invoke_transport() function
			 */
			USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_transport(12) End with Fail !! \n")) ;
			return USB_STOR_TRANSPORT_ERROR;
   }

 /* we should never get here, but if we do, we're in trouble */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_transport(13) End with Fail !! \n")) ;
 return USB_STOR_TRANSPORT_ERROR;
}

/***********************************************************************
 * Reset routines
 ***********************************************************************/

/* This is the common part of the device reset code.
 *
 * It's handy that every transport mechanism uses the control endpoint for
 * resets.
 *
 * Basically, we send a reset with a 20-second timeout, so we don't get
 * jammed attempting to do the reset.
 */
static int usb_stor_reset_common(struct us_data *us,
		u8 request, u8 requesttype,
		u16 value, u16 index, void *data, u16 size)
{
 int result;
 int result2;

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_reset_common() Begin ..\n"));
 /* A 20-second timeout may seem rather long, but a LaCie
  *  StudioDrive USB2 device takes 16+ seconds to get going
  *  following a powerup or USB attach event. */

 #ifdef USBMS_BS_ASSERT
	if (data) {
		if ( !(USB_IS_UN_CACHED(data)) ) {
				iprintf("[ASSERT][USBMS] usb_stor_reset_common() : " ) ;
	   		    iprintf("Error : Not Un-Cached !!\n" ) ;
			    iprintf("[ASSERT][USBMS] data addr = %08XH \n", data ) ;
			    iprintf("[ASSERT][USBMS] Continue ..\n") ;
   			   }
	   }
 #endif // USBMS_BS_ASSERT
 
 /* Use usb_control_msg() because this is not a queued-command */
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_reset_common() : usb_stor_control_msg() ..\n"));
 result = usb_stor_control_msg(us, us->send_ctrl_pipe,
			request, requesttype, value, index, data, size,
			/*
			20*HZ);
			*/
			USBMS_RESET_TIMEOUT );

	if (result < 0) {
		US_DEBUGP("Soft reset failed: %d\n", result);
		return FAILED;
	}

	/* long wait for reset, so unlock to allow disconnects */
 #ifndef NO_SEMPHORE
	up(&us->dev_semaphore);
 #endif // ndef NO_SEMPHORE
	/*
    #ifndef _ETCPIP
	set_current_state(TASK_UNINTERRUPTIBLE);
	schedule_timeout(HZ*6);
    #endif // ndef _ETCPIP
	*/
 #ifndef NO_SEMPHORE
	down(&us->dev_semaphore);
 #endif // ndef NO_SEMPHORE
	if (test_bit(US_FLIDX_DISCONNECTING, &us->flags)) {
		US_DEBUGP("Reset interrupted by disconnect\n");
		return FAILED;
	}

	US_DEBUGP("Soft reset: clearing bulk-in endpoint halt\n");
	result = usb_stor_clear_halt(us, us->recv_bulk_pipe);

	US_DEBUGP("Soft reset: clearing bulk-out endpoint halt\n");
	result2 = usb_stor_clear_halt(us, us->send_bulk_pipe);

	/* return a result code based on the result of the control message */
	if (result < 0 || result2 < 0) {
		US_DEBUGP("Soft reset failed\n");
		return FAILED;
	}
		US_DEBUGP("Soft reset done\n");
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_reset_common() End.\n"));
	return SCSI_SUCCESS;
}

/* This issues a CB[I] Reset to the device in question
 */
#define CB_RESET_CMD_SIZE	12

int usb_stor_CB_reset(struct us_data *us)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_CB_reset() Begin .. \n"));
 US_DEBUGP("%s called\n", __FUNCTION__);

 memset(us->iobuf, 0xFF, CB_RESET_CMD_SIZE);
 us->iobuf[0] = SEND_DIAGNOSTIC;
 us->iobuf[1] = 4;
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_CB_reset() End with Call usb_stor_reset_common() .. \n"));
 return usb_stor_reset_common(us, US_CBI_ADSC, 
				 USB_TYPE_CLASS | USB_RECIP_INTERFACE,
				 0, us->ifnum, us->iobuf, CB_RESET_CMD_SIZE);
}

/* This issues a Bulk-only Reset to the device in question, including
 * clearing the subsequent endpoint halts that may occur.
 */
int usb_stor_Bulk_reset(struct us_data *us)
{
 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_reset() Begin .. \n")) ;
 US_DEBUGP("%s called\n", __FUNCTION__);

 USBMS_DBG_FLOW(iprintf("[USBMS FLOW] usb_stor_Bulk_reset() End with Call usb_stor_reset_common() ..\n"));
 return usb_stor_reset_common(us, US_BULK_RESET_REQUEST, 
				 USB_TYPE_CLASS | USB_RECIP_INTERFACE,
				 0, us->ifnum, NULL, 0);
}

// B. S. Liaw TEST for compiler warning !!

